/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/time.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/queue.h>
#include <sys/tree.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>

#include <event.h>
#include <evhttp.h>

#include "spybye2html.h"
#include "spybye.gen.h"
#include "log.h"

static struct event ev_timeout;
static char *input = NULL;
static char *output = NULL;

void
print_header(FILE *outp)
{
	fprintf(outp, "<html><head>");
	fprintf(outp,
	    "<style type=\"text/css\">"
	    "<!--"
	    "table.sites {"
	    "  border-width: 0px 0px 1px 1px;"
	    "  border-spacing: 2px;"
	    "  border-style: inset;"
	    "  border-color: black;"
	    "  border-collapse: collapse;"
	    "}"
	    "table.sites td {"
	    "  border-width: 1px 1px 0px 0px;"
	    "  padding: 2px;"
	    "  border-style: inset;"
	    "  border-color: black;"
	    "  background-color: rgb(255, 250, 220);"
	    "  font-family: Verdana, Arial;\n"
	    "  font-size: small;\n"
	    "}\n"
	    "table.sites tr {"
	    "  border-width: 1px 1px 0px 0px;"
	    "  padding: 2px;"
	    "  border-style: inset;"
	    "  border-color: black;"
	    "  background-color: rgb(235, 170, 200);"
	    "  font-family: Verdana, Arial;\n"
	    "  font-size: 0.9em;\n"
	    "}\n"
	    "--></style></head>");

	fprintf(outp,
	    "<body><table class=sites>"
	    "<tr>"
	    "<th>Report Time</th><th>Origin URL</th><th>Danger URL</th>"
	    "</tr>"
	    );
}

void
print_footer(FILE *outp)
{
	fprintf(outp, "</table></body></html>");
}

void
format_dangerousload(FILE *outp, struct dangerousload *dl)
{
	char output[64];
	struct tm *tm;
	u_int tmp;
	time_t seconds;
	char *parent_url, *danger_url, *escaped;

	EVTAG_GET(dl, time_in_seconds, &tmp);
	EVTAG_GET(dl, parent_url, &parent_url);
	EVTAG_GET(dl, dangerous_url, &danger_url);

	seconds = tmp;
	tm = localtime(&seconds);
	strftime(output, sizeof(output), "%Y-%m-%d %H:%M:%S", tm);

	fprintf(outp, "<tr>");
	fprintf(outp, "<td><span class=time>%s</span></td>", output);
	escaped = evhttp_htmlescape(parent_url);
	fprintf(outp, "<td><span class=source>%s</span></td>", escaped);
	free(escaped);
	escaped = evhttp_htmlescape(danger_url);
	fprintf(outp, "<td><span class=danger>%s</span></td>", escaped);
	free(escaped);
	fprintf(outp, "</tr>\n");
}

void
format_dangerous_report(FILE *outp, struct dangerous_report *dr)
{
	struct dangerousload *dl;

	EVTAG_GET(dr, report, &dl);

	format_dangerousload(outp, dl);
}

void
transform(FILE *inp, FILE *outp)
{
	struct evbuffer *data = evbuffer_new();
	int count;
	assert(data != NULL);

	print_header(outp);

	count = 0;
	while (!feof(inp)) {
		u_int32_t length;
		char buffer[1024];
		size_t n = fread(buffer, 1, 1024, inp);
		/* why would this be - oh no */
		if (n == 0)
			continue;
		evbuffer_add(data, buffer, n);

		while (evtag_peek_length(data, &length) != -1) {
			u_char tag;

			if (EVBUFFER_LENGTH(data) < length)
				break;

			evtag_peek(data, &tag);

			count++;
			
			if (tag == DANGEROUS_REPORT_TAG) {	
				struct dangerous_report *dr;
				if ((dr = dangerous_report_new()) == NULL)
					err(1, "malloc");

				if (evtag_unmarshal_dangerous_report(data,
					DANGEROUS_REPORT_TAG, dr) == -1) {
					fprintf(stderr,
					    "Failed to read entry %d\n",
					    count);
					dangerous_report_free(dr);
					continue;
				}

				format_dangerous_report(outp, dr);
				
				dangerous_report_free(dr);
			} else if (tag == DANGEROUS_TAG) {
				struct dangerousload *dl;
				if ((dl = dangerousload_new()) == NULL)
					err(1, "malloc");

				if (evtag_unmarshal_dangerousload(data,
					DANGEROUS_TAG, dl) == -1) {
					fprintf(stderr,
					    "Failed to read entry %d\n",
					    count);
					dangerousload_free(dl);
					continue;
				}

				format_dangerousload(outp, dl);
				
				dangerousload_free(dl);
			}
		}
	}

	print_footer(outp);

	fprintf(stderr, " ... read %d entries\n", count);

	evbuffer_free(data);
}

void
transform_cb(int fd, short what, void *arg)
{
	struct timeval tv;
	FILE *inputp = fopen(input, "r");
	FILE *outputp = fopen(output, "w+");

	timerclear(&tv);
	tv.tv_sec = TIME_INTERVAL;
	event_add(&ev_timeout, &tv);

	if (inputp == NULL)
		goto done;
	if (outputp == NULL)
		goto done;

	fprintf(stderr, "Reading data from %s to %s\n",
	    input, output);

	transform(inputp, outputp);

	fclose(outputp);
	fclose(inputp);

done:
	if (inputp != NULL)
		fclose(inputp);
	if (outputp != NULL)
		fclose(outputp);
}

void
usage(const char *progname)
{
	fprintf(stderr,
	    "%s: [-i input] [-o output]\n"
	    "\t -i input file\n"
	    "\t -o output file\n"
	    "\t for documentation of all options consult the man page\n",
	    progname);
}

int
main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	int ch;
	struct timeval tv;

	while ((ch = getopt(argc, argv, "i:o:")) != -1)
		switch(ch) {
		case 'i':
			input = optarg;
			break;
		case 'o':
			output = optarg;
			break;

		default:
			usage(argv[0]);
			exit(1);
		}

	event_init();
	evtag_init();

	if (input == NULL || output == NULL) {
		FILE *inputp = input ? fopen(input, "r") : stdin;
		FILE *outputp = output ? fopen(output, "w") : stdout;
		if (inputp == NULL) {
			fprintf(stderr, "Cannot open %s\n", input);
			usage(argv[0]);
			exit(1);
		}
		if (outputp == NULL) {
			fprintf(stderr, "Cannot open %s\n", output);
			usage(argv[0]);
			exit(1);
		}
		transform(inputp, outputp);
		exit(0);
	}

	argc -= optind;
	argv += optind;

	fprintf(stderr, "spybye2html %s starting up ...\n", VERSION);

	evtimer_set(&ev_timeout, transform_cb, NULL);
	timerclear(&tv);
	event_add(&ev_timeout, &tv);
	
	event_dispatch();

	exit (0);
}
